from typing import Dict, List, Optional
import httpx
import nacl.signing
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Confirmed
from solana.rpc.types import TxOpts
from solders.transaction import Transaction
from solders.instruction import Instruction
from solders.keypair import Keypair
from solders.message import Message
from solders.signature import Signature
from solders.pubkey import Pubkey


class SolanaTransaction:
    """Transaction parameters for Solana."""

    def __init__(
        self,
        instructions: List[Instruction],
        accounts_to_sign: Optional[List[Keypair]] = None,
    ):
        self.instructions = instructions
        self.accounts_to_sign = accounts_to_sign


class SolanaWalletClient:
    """Solana wallet implementation."""

    def __init__(
        self,
        rpc_url: str,
        keypair: Optional[Keypair] = None,
        pubkey: Optional[str] = None,
        fee_payer: Optional[str] = None,
    ):
        self.client = AsyncClient(rpc_url)
        self.rpc_url = rpc_url
        self.keypair = keypair
        self.pubkey = pubkey
        if pubkey:
            self.pubkey = Pubkey.from_string(pubkey)
        elif keypair:
            self.pubkey = keypair.pubkey()
        if fee_payer:
            self.fee_payer = Keypair.from_base58_string(fee_payer)

    def sign_message(self, message: bytes) -> Signature:
        signed = nacl.signing.SigningKey(self.keypair.secret()).sign(message)
        return Signature.from_bytes(signed.signature)

    def send_transaction(self, transaction: SolanaTransaction) -> Dict[str, str]:
        recent_blockhash = self.client.get_latest_blockhash().value.blockhash
        payer = self.keypair.pubkey()

        ixs = []
        for instruction in transaction.instructions:
            ixs.append(instruction)

        signers = [self.keypair]
        if transaction.accounts_to_sign:
            signers.extend(transaction.accounts_to_sign)

        message = Message(
            instructions=ixs,
            recent_blockhash=recent_blockhash,
            payer=payer,
        )
        tx = Transaction(
            from_keypairs=signers,
            message=message,
            recent_blockhash=recent_blockhash,
        )
        result = self.client.send_transaction(
            tx,
            opts=TxOpts(
                skip_preflight=False, max_retries=10, preflight_commitment=Confirmed
            ),
        )
        self.client.confirm_transaction(result.value, commitment=Confirmed)
        return {"hash": str(result.value)}

    async def get_priority_fee_estimate_helius(
        self,
        serialized_transaction: str,
    ) -> int:
        """
        Get the priority fee estimate from Helius.

        :param serialized_transaction: The base64-encoded serialized transaction.
        :return: The estimated priority fee (int).
        """
        payload = {
            "jsonrpc": "2.0",
            "id": "1",
            "method": "getPriorityFeeEstimate",
            "params": [
                {
                    "transaction": serialized_transaction,
                    "options": {"recommended": True},
                }
            ],
        }
        async with httpx.AsyncClient() as client:
            response = await client.post(self.rpc_url, json=payload)
            response.raise_for_status()
            result = response.json()
            if "error" in result:
                raise RuntimeError(f"Fee estimation failed: {result['error']}")
            return int(result["result"]["priorityFeeEstimate"])



import asyncio
import logging
from typing import Dict, Any, List, Optional
from solana_agent import AutoTool, ToolRegistry
import httpx


def summarize_alphavybe_balances(api_response: Dict[str, Any]) -> str:
    lines = []
    sol = api_response.get("solBalance")
    sol_usd = api_response.get("solBalanceUsd")
    staked_sol = api_response.get("stakedSolBalance")
    staked_sol_usd = api_response.get("stakedSolBalanceUsd")
    active_staked_sol = api_response.get("activeStakedSolBalance")
    active_staked_sol_usd = api_response.get("activeStakedSolBalanceUsd")
    total_usd = api_response.get("totalTokenValueUsd")
    total_usd_change = api_response.get("totalTokenValueUsd1dChange")
    owner = api_response.get("ownerAddress")

    lines.append(f"Wallet: {owner}")
    if sol is not None:
        lines.append(f"SOL balance: {sol} (~${sol_usd})")
    if staked_sol is not None:
        lines.append(f"Staked SOL: {staked_sol} (~${staked_sol_usd})")
    if active_staked_sol is not None:
        lines.append(
            f"Active Staked SOL: {active_staked_sol} (~${active_staked_sol_usd})"
        )
    if total_usd is not None:
        lines.append(
            f"Total wallet value: ${total_usd} (24h change: {total_usd_change})"
        )

    tokens = api_response.get("data", [])
    if tokens:
        lines.append("Top tokens:")
        for token in tokens[:10]:
            name = token.get("name") or token.get("symbol") or token.get("mintAddress")
            amount = token.get("amount")
            symbol = token.get("symbol") or ""
            value_usd = token.get("valueUsd")
            verified = "✅" if token.get("verified") else "❌"
            lines.append(f"  - {name} ({symbol}): {amount} (${value_usd}) {verified}")
    else:
        lines.append("No SPL tokens found.")
    return "\n".join(lines)


async def get_privy_embedded_wallet_address(
    user_id: str, app_id: str, app_secret: str
) -> Optional[str]:
    url = f"https://auth.privy.io/api/v1/users/{user_id}"
    headers = {"privy-app-id": app_id}
    auth = (app_id, app_secret)
    async with httpx.AsyncClient() as client:
        resp = await client.get(url, headers=headers, auth=auth, timeout=10)
        resp.raise_for_status()
        data = resp.json()
        for acct in data.get("linked_accounts", []):
            if acct.get("connector_type") == "embedded" and acct.get("delegated"):
                return acct["public_key"]
    return None


class PrivyBalanceCheckerTool(AutoTool):
    def __init__(self, registry: Optional[ToolRegistry] = None):
        super().__init__(
            name="privy_balance",
            description="Check SOL and SPL token balances for a Privy delegated embedded wallet using AlphaVybe API.",
            registry=registry,
        )
        self.api_key = None
        self.app_id = None
        self.app_secret = None

    def get_schema(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "user_id": {
                    "type": "string",
                    "description": "Privy user id (did) to check delegated embedded wallet balance.",
                }
            },
            "required": ["user_id"],
            "additionalProperties": False,
        }

    def configure(self, config: Dict[str, Any]) -> None:
        tool_cfg = config.get("tools", {}).get("privy_balance", {})
        self.api_key = tool_cfg.get("api_key")
        self.app_id = tool_cfg.get("app_id")
        self.app_secret = tool_cfg.get("app_secret")

    async def execute(self, user_id: str) -> Dict[str, Any]:
        if not all([self.api_key, self.app_id, self.app_secret]):
            return {"status": "error", "message": "Privy or AlphaVybe config missing."}
        wallet_address = await get_privy_embedded_wallet_address(
            user_id, self.app_id, self.app_secret
        )
        if not wallet_address:
            return {
                "status": "error",
                "message": "No delegated embedded wallet found for user.",
            }
        url = f"https://api.vybenetwork.xyz/account/token-balance/{wallet_address}?limit=10&sortByDesc=valueUsd"
        headers = {
            "accept": "application/json",
            "X-API-KEY": self.api_key,
        }
        max_retries = 3
        for attempt in range(max_retries):
            try:
                async with httpx.AsyncClient(timeout=10.0) as client:
                    resp = await client.get(url, headers=headers, timeout=15)
                    if resp.status_code >= 500:
                        raise httpx.HTTPStatusError("Server error", request=resp.request, response=resp)
                    resp.raise_for_status()
                    data = resp.json()
                    summary = summarize_alphavybe_balances(data)
                    return {
                        "status": "success",
                        "result": summary,
                    }
            except httpx.HTTPStatusError as e:
                if resp.status_code >= 500 and attempt < max_retries - 1:
                    await asyncio.sleep(1)
                    continue
                logging.exception(f"Privy balance check error: {e}")
                return {"status": "error", "message": str(e)}
            except Exception as e:
                logging.exception(f"Privy balance check error: {e}")
                return {"status": "error", "message": str(e)}

class PrivyBalanceCheckerPlugin:
    def __init__(self):
        self.name = "privy_balance"
        self.config = None
        self.tool_registry = None
        self._tool = None

    @property
    def description(self):
        return "Plugin for checking SOL and SPL token balances for a Privy delegated embedded wallet using AlphaVybe API."

    def initialize(self, tool_registry: ToolRegistry) -> None:
        self.tool_registry = tool_registry
        self._tool = PrivyBalanceCheckerTool(registry=tool_registry)

    def configure(self, config: Dict[str, Any]) -> None:
        self.config = config
        if self._tool:
            self._tool.configure(self.config)

    def get_tools(self) -> List[AutoTool]:
        return [self._tool] if self._tool else []


def get_plugin():
    return PrivyBalanceCheckerPlugin()



from typing import Dict, Any, List, Optional
from solana_agent import AutoTool, ToolRegistry
import httpx

class SnsLookupTool(AutoTool):
    def __init__(self, registry: Optional[ToolRegistry] = None):
        super().__init__(
            name="sns_lookup",
            description="Lookup SNS domain with a QuickNode RPC.",
            registry=registry,
        )
        self.quicknode_url = None

    def get_schema(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "domain": {
                    "type": "string",
                    "description": "SNS domain to resolve, e.g. domain.sol",
                }
            },
            "required": ["domain"],
            "additionalProperties": False,
        }

    def configure(self, config: Dict[str, Any]) -> None:
        super().configure(config)
        tool_cfg = config.get("tools", {}).get("sns_lookup", {})
        self.quicknode_url = tool_cfg.get("quicknode_url")

    async def execute(self, domain: str) -> Dict[str, Any]:
        if not self.quicknode_url:
            return {"status": "error", "message": "No quicknode_url configured for SNS lookup."}
        payload = {
            "method": "sns_resolveDomain",
            "params": [domain],
            "id": 1,
            "jsonrpc": "2.0",
        }
        headers = {"Content-Type": "application/json"}
        async with httpx.AsyncClient() as client:
            resp = await client.post(self.quicknode_url, json=payload, headers=headers)
            resp.raise_for_status()
            data = resp.json()
            return {"status": "success", "result": data}

class SnsLookupPlugin:
    def __init__(self):
        self.name = "sns_lookup"
        self.config = None
        self.tool_registry = None
        self._tool = None

    @property
    def description(self):
        return "Plugin for resolving SNS domains using a QuickNode endpoint."

    def initialize(self, tool_registry: ToolRegistry) -> None:
        self.tool_registry = tool_registry
        self._tool = SnsLookupTool(registry=tool_registry)

    def configure(self, config: Dict[str, Any]) -> None:
        self.config = config
        if self._tool:
            self._tool.configure(self.config)

    def get_tools(self) -> List[AutoTool]:
        return [self._tool] if self._tool else []

def get_plugin():
    return SnsLookupPlugin()


import base64
import json
import logging
from typing import Dict, Any, List, Optional
from solana_agent import AutoTool, ToolRegistry
import httpx
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from sakit.utils.wallet import SolanaWalletClient
from sakit.utils.transfer import TokenTransferManager

LAMPORTS_PER_SOL = 10**9
SPL_TOKEN_PROGRAM_ID = "TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"
TOKEN_2022_PROGRAM_ID = "TokenzQdBNbLqP5VEhdkAS6EPFLC1PHnBqCXEpPxuEb"


def canonicalize(obj):
    return json.dumps(obj, sort_keys=True, separators=(",", ":"))


def get_authorization_signature(url, body, privy_app_id, privy_auth_key):
    payload = {
        "version": 1,
        "method": "POST",
        "url": url,
        "body": body,
        "headers": {"privy-app-id": privy_app_id},
    }
    serialized_payload = canonicalize(payload)
    private_key_string = privy_auth_key.replace("wallet-auth:", "")
    private_key_pem = (
        f"-----BEGIN PRIVATE KEY-----\n{private_key_string}\n-----END PRIVATE KEY-----"
    )
    private_key = serialization.load_pem_private_key(
        private_key_pem.encode("utf-8"), password=None
    )
    signature = private_key.sign(
        serialized_payload.encode("utf-8"), ec.ECDSA(hashes.SHA256())
    )
    return base64.b64encode(signature).decode("utf-8")


async def get_privy_embedded_wallet(
    user_id: str, app_id: str, app_secret: str
) -> Optional[Dict[str, str]]:
    url = f"https://auth.privy.io/api/v1/users/{user_id}"
    headers = {"privy-app-id": app_id}
    auth = (app_id, app_secret)
    async with httpx.AsyncClient() as client:
        resp = await client.get(url, headers=headers, auth=auth, timeout=10)
        if resp.status_code != 200:
            logging.error(f"Privy API error: {resp.text}")
            resp.raise_for_status()
        data = resp.json()
        for acct in data.get("linked_accounts", []):
            if acct.get("connector_type") == "embedded" and acct.get("delegated"):
                return {"wallet_id": acct["id"], "public_key": acct["public_key"]}
    return None


async def privy_sign_and_send(
    wallet_id: str, encoded_tx: str, app_id: str, app_secret: str, privy_auth_key: str
) -> Dict[str, Any]:
    url = f"https://api.privy.io/v1/wallets/{wallet_id}/rpc"
    auth_string = f"{app_id}:{app_secret}"
    encoded_auth = base64.b64encode(auth_string.encode()).decode()
    body = {
        "method": "signAndSendTransaction",
        "caip2": "solana:5eykt4UsFv8P8NJdTREpY1vzqKqZKvdp",
        "params": {"transaction": encoded_tx, "encoding": "base64"},
    }
    signature = get_authorization_signature(
        url=url, body=body, privy_app_id=app_id, privy_auth_key=privy_auth_key
    )
    headers = {
        "Authorization": f"Basic {encoded_auth}",
        "privy-app-id": app_id,
        "privy-authorization-signature": signature,
        "Content-Type": "application/json",
    }
    async with httpx.AsyncClient() as client:
        resp = await client.post(url, headers=headers, json=body, timeout=10)
        if resp.status_code != 200:
            logging.error(f"Privy sign and send failed: {resp.text}")
            resp.raise_for_status()
        return resp.json()


class PrivyTransferTool(AutoTool):
    def __init__(self, registry: Optional[ToolRegistry] = None):
        super().__init__(
            name="privy_transfer",
            description="Transfer SOL or SPL tokens using Privy delegated wallet.",
            registry=registry,
        )
        self.app_id = None
        self.app_secret = None
        self.signing_key = None
        self.rpc_url = None
        self.fee_payer = None
        self.fee_percentage = None

    def get_schema(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "user_id": {"type": "string", "description": "Privy user id (did)"},
                "to_address": {
                    "type": "string",
                    "description": "Recipient wallet address",
                },
                "amount": {
                    "type": "number",
                    "description": "Amount to transfer (in SOL or token units)",
                },
                "mint": {
                    "type": "string",
                    "description": "Token mint address",
                },
                "memo": {
                    "type": "string",
                    "description": "Optional memo for the transaction",
                    "default": "",
                },
            },
            "required": ["user_id", "to_address", "amount", "mint", "memo"],
            "additionalProperties": False,
        }

    def configure(self, config: Dict[str, Any]) -> None:
        tool_cfg = config.get("tools", {}).get("privy_transfer", {})
        self.app_id = tool_cfg.get("app_id")
        self.app_secret = tool_cfg.get("app_secret")
        self.signing_key = tool_cfg.get("signing_key")
        self.rpc_url = tool_cfg.get("rpc_url")
        self.fee_payer = tool_cfg.get("fee_payer")
        self.fee_percentage = tool_cfg.get("fee_percentage", 0.85)  # Default 0.85% fee

    async def execute(
        self,
        user_id: str,
        to_address: str,
        amount: float,
        mint: str,
        memo: str = "",
    ) -> Dict[str, Any]:
        if not all(
            [
                self.app_id,
                self.app_secret,
                self.signing_key,
                self.rpc_url,
                self.fee_payer,
            ]
        ):
            return {"status": "error", "message": "Privy config missing."}
        wallet_info = await get_privy_embedded_wallet(
            user_id, self.app_id, self.app_secret
        )
        if not wallet_info:
            return {
                "status": "error",
                "message": "No delegated embedded wallet found for user.",
            }
        wallet_id = wallet_info["wallet_id"]
        try:
            wallet = SolanaWalletClient(
                self.rpc_url, None, wallet_info["public_key"], self.fee_payer
            )
            provider = None
            if "helius" in self.rpc_url:
                provider = "helius"
            transaction = await TokenTransferManager.transfer(
                wallet, to_address, amount, mint, provider, True, self.fee_percentage, memo
            )
            encoded_transaction = base64.b64encode(bytes(transaction)).decode("utf-8")
            result = await privy_sign_and_send(
                wallet_id,
                encoded_transaction,
                self.app_id,
                self.app_secret,
                self.signing_key,
            )
            return {"status": "success", "result": result}
        except Exception as e:
            logging.exception(f"Privy transfer failed: {str(e)}")
            return {"status": "error", "message": str(e)}


class PrivyTransferPlugin:
    def __init__(self):
        self.name = "privy_transfer"
        self.config = None
        self.tool_registry = None
        self._tool = None

    @property
    def description(self):
        return "Plugin for transferring SOL or SPL tokens using Privy delegated wallet."

    def initialize(self, tool_registry: ToolRegistry) -> None:
        self.tool_registry = tool_registry
        self._tool = PrivyTransferTool(registry=tool_registry)

    def configure(self, config: Dict[str, Any]) -> None:
        self.config = config
        if self._tool:
            self._tool.configure(self.config)

    def get_tools(self) -> List[AutoTool]:
        return [self._tool] if self._tool else []


def get_plugin():
    return PrivyTransferPlugin()



import logging
from typing import Dict, Any, List, Optional
from solana_agent import AutoTool, ToolRegistry
import httpx


def summarize_birdeye_price(api_response: Dict[str, Any], address: str) -> str:
    if not api_response.get("success"):
        return f"Failed to fetch price for {address}."
    data = api_response.get("data", {})
    value = data.get("value")
    price_change = data.get("priceChange24h")
    liquidity = data.get("liquidity")
    update_time = data.get("updateHumanTime")
    lines = [
        f"Token address: {address}",
        f"Current price: {value}",
        f"24h price change: {price_change}%",
        f"Liquidity: {liquidity}",
        f"Last updated: {update_time}",
    ]
    return "\n".join(lines)


class BirdeyePriceCheckerTool(AutoTool):
    def __init__(self, registry: Optional[ToolRegistry] = None):
        super().__init__(
            name="solana_price",
            description="Get the current price and liquidity for a Solana token using Birdeye API.",
            registry=registry,
        )
        self.api_key = None

    def get_schema(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "address": {
                    "type": "string",
                    "description": "The SPL token mint address to check price for.",
                }
            },
            "required": ["address"],
            "additionalProperties": False,
        }

    def configure(self, config: Dict[str, Any]) -> None:
        super().configure(config)
        tool_cfg = config.get("tools", {}).get("solana_price", {})
        self.api_key = tool_cfg.get("api_key")

    async def execute(self, address: str) -> Dict[str, Any]:
        if not self.api_key:
            return {"status": "error", "message": "Birdeye API key not configured."}
        url = f"https://public-api.birdeye.so/defi/price?include_liquidity=true&address={address}"
        headers = {
            "accept": "application/json",
            "x-chain": "solana",
            "X-API-KEY": self.api_key,
        }
        try:
            async with httpx.AsyncClient(timeout=10.0) as client:
                resp = await client.get(url, headers=headers)
                resp.raise_for_status()
                data = resp.json()
                summary = summarize_birdeye_price(data, address)
                return {
                    "status": "success",
                    "result": summary,
                }
        except Exception as e:
            logging.exception(f"Birdeye price check error: {e}")
            return {"status": "error", "message": str(e)}


class BirdeyePriceCheckerPlugin:
    def __init__(self):
        self.name = "solana_price"
        self.config = None
        self.tool_registry = None
        self._tool = None

    @property
    def description(self):
        return "Plugin for checking Solana token price and liquidity using Birdeye API."

    def initialize(self, tool_registry: ToolRegistry) -> None:
        self.tool_registry = tool_registry
        self._tool = BirdeyePriceCheckerTool(registry=tool_registry)

    def configure(self, config: Dict[str, Any]) -> None:
        self.config = config
        if self._tool:
            self._tool.configure(self.config)

    def get_tools(self) -> List[AutoTool]:
        return [self._tool] if self._tool else []


def get_plugin():
    return BirdeyePriceCheckerPlugin()



import logging
import base64
from typing import Optional, List
import httpx

from solders.pubkey import Pubkey
from solana.rpc.commitment import Finalized
from solders.transaction import VersionedTransaction
from solders.message import (
    to_bytes_versioned,
    MessageV0,
)
from spl.token.instructions import (
    transfer_checked as spl_transfer,
    TransferCheckedParams as SPLTransferParams,
)
from solders.address_lookup_table_account import AddressLookupTableAccount
from solders.null_signer import NullSigner
from solders.instruction import Instruction, AccountMeta
from solders.system_program import TransferParams, transfer
from solders.compute_budget import set_compute_unit_limit, set_compute_unit_price
from solders.signature import Signature
from sakit.utils.wallet import SolanaWalletClient

JUP_API = "https://quote-api.jup.ag/v6"
SPL_TOKEN_PROGRAM_ID = "TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"
TOKEN_2022_PROGRAM_ID = "TokenzQdBNbLqP5VEhdkAS6EPFLC1PHnBqCXEpPxuEb"
LAMPORTS_PER_SOL = 10**9

class TradeManager:
    @staticmethod
    def parse_address_table_lookups(addresses):
        if not addresses:
            return []
        if isinstance(addresses, str):
            addresses = [addresses]
        # Use empty addresses for simulation/compilation
        return [AddressLookupTableAccount(Pubkey.from_string(addr), []) for addr in addresses]

    @staticmethod
    def parse_instruction(ix_obj) -> Instruction:
        program_id = Pubkey.from_string(ix_obj["programId"])
        accounts = [
            AccountMeta(
                pubkey=Pubkey.from_string(acc["pubkey"]),
                is_signer=acc["isSigner"],
                is_writable=acc["isWritable"],
            )
            for acc in ix_obj["accounts"]
        ]
        data = base64.b64decode(ix_obj["data"])
        return Instruction(program_id=program_id, accounts=accounts, data=data)

    @staticmethod
    def parse_instruction_list(ix_list) -> List[Instruction]:
        return [TradeManager.parse_instruction(ix) for ix in ix_list]

    @staticmethod
    async def trade(
        wallet: SolanaWalletClient,
        output_mint: str,
        input_amount: float,
        input_mint: str = None,
        slippage_bps: int = 300,
        jupiter_url: Optional[str] = None,
        no_signer: bool = False,
        provider: Optional[str] = None,
        fee_percentage: float = 0.85,
    ) -> VersionedTransaction:
        """
        Swap tokens using Jupiter Exchange, with compute budget and optional priority fee (if provider == 'helius').
        """
        try:
            if input_mint == "So11111111111111111111111111111111111111112":
                adjusted_amount = int(input_amount * LAMPORTS_PER_SOL)
            else:
                mint_pubkey = Pubkey.from_string(input_mint)
                resp = await wallet.client.get_account_info(mint_pubkey)
                owner = str(resp.value.owner)
                if owner == SPL_TOKEN_PROGRAM_ID:
                    program_id = Pubkey.from_string(SPL_TOKEN_PROGRAM_ID)
                elif owner == TOKEN_2022_PROGRAM_ID:
                    program_id = Pubkey.from_string(TOKEN_2022_PROGRAM_ID)
                else:
                    raise ValueError(
                        f"Unsupported token program: {owner}. Supported programs are SPL Token and Token 2022."
                    )
                from spl.token.async_client import AsyncToken
                token = AsyncToken(wallet.client, mint_pubkey, program_id, wallet.fee_payer)
                mint_info = await token.get_mint_info()
                adjusted_amount = int(input_amount * (10**mint_info.decimals))

            if jupiter_url is None:
                jupiter_url = JUP_API

            quote_url = (
                f"{jupiter_url}/quote?"
                f"inputMint={input_mint}"
                f"&outputMint={output_mint}"
                f"&amount={adjusted_amount}"
                f"&slippageBps={slippage_bps}"
                f"&onlyDirectRoutes=true"
                f"&maxAccounts=20"
            )

            async with httpx.AsyncClient() as client:
                quote_response = await client.get(quote_url)
                if quote_response.status_code != 200:
                    raise Exception(
                        f"Failed to fetch quote: {quote_response.status_code}"
                    )
                quote_data = quote_response.json()

                swap_instructions_response = await client.post(
                    f"{jupiter_url}/swap-instructions",
                    json={
                        "quoteResponse": quote_data,
                        "userPublicKey": str(wallet.pubkey),
                        "wrapAndUnwrapSol": True,
                        "dynamicComputeUnitLimit": True,
                        "prioritizationFeeLamports": "auto",
                    },
                )
                if swap_instructions_response.status_code != 200:
                    raise Exception(
                        f"Failed to fetch swap instructions: {swap_instructions_response.status_code}"
                    )
                swap_instructions_data = swap_instructions_response.json()

            # Build all instructions in order
            instructions = []

            # 1. Compute budget instructions (if present)
            has_cu = (
                "computeBudgetInstructions" in swap_instructions_data
                and swap_instructions_data["computeBudgetInstructions"]
            )
            if has_cu:
                instructions += TradeManager.parse_instruction_list(swap_instructions_data["computeBudgetInstructions"])

            # 2. Setup instructions (if present)
            if "setupInstructions" in swap_instructions_data and swap_instructions_data["setupInstructions"]:
                instructions += TradeManager.parse_instruction_list(swap_instructions_data["setupInstructions"])

            # 3. Fee instruction (if wallet.fee_payer)
            if wallet.fee_payer:
                if input_mint == "So11111111111111111111111111111111111111112":
                    ix_fee = transfer(
                        TransferParams(
                            from_pubkey=wallet.pubkey,
                            to_pubkey=wallet.fee_payer.pubkey(),
                            lamports=int(input_amount * LAMPORTS_PER_SOL * (fee_percentage / 100)),
                        )
                    )
                    instructions.append(ix_fee)
                else:
                    mint_pubkey = Pubkey.from_string(input_mint)
                    resp = await wallet.client.get_account_info(mint_pubkey)
                    owner = str(resp.value.owner)
                    if owner == SPL_TOKEN_PROGRAM_ID:
                        program_id = Pubkey.from_string(SPL_TOKEN_PROGRAM_ID)
                    elif owner == TOKEN_2022_PROGRAM_ID:
                        program_id = Pubkey.from_string(TOKEN_2022_PROGRAM_ID)
                    else:
                        raise ValueError(
                            f"Unsupported token program: {owner}. Supported programs are SPL Token and Token 2022."
                        )

                    token = AsyncToken(
                        wallet.client, mint_pubkey, program_id, wallet.fee_payer
                    )

                    from_ata = (
                        (await token.get_accounts_by_owner(wallet.pubkey)).value[0].pubkey
                    )
                    mint_info = await token.get_mint_info()
                    adjusted_amount = int(input_amount * (10**mint_info.decimals))

                    to_fee_ata = (await token.get_accounts_by_owner(wallet.fee_payer.pubkey())).value[0].pubkey
                    fee_amount = int(adjusted_amount * (fee_percentage / 100))
                    ix_fee = spl_transfer(
                        SPLTransferParams(
                            program_id=program_id,
                            source=from_ata,
                            mint=mint_pubkey,
                            dest=to_fee_ata,
                            owner=wallet.pubkey,
                            amount=fee_amount,
                            decimals=mint_info.decimals,
                        )
                    )
                    instructions.append(ix_fee)

                    webhook_fee = transfer(
                        TransferParams(
                            from_pubkey=wallet.pubkey,
                            to_pubkey=wallet.fee_payer.pubkey(),
                            lamports=int(0.0001 * LAMPORTS_PER_SOL),
                        )
                    )
                    instructions.append(webhook_fee)

            # 4. Swap instruction (required)
            swap_instruction = TradeManager.parse_instruction(swap_instructions_data["swapInstruction"])
            instructions.append(swap_instruction)

            # 5. Cleanup instruction (if present)
            if "cleanupInstruction" in swap_instructions_data and swap_instructions_data["cleanupInstruction"]:
                instructions.append(TradeManager.parse_instruction(swap_instructions_data["cleanupInstruction"]))

            # 6. Other instructions (if present)
            if "otherInstructions" in swap_instructions_data and swap_instructions_data["otherInstructions"]:
                instructions += TradeManager.parse_instruction_list(swap_instructions_data["otherInstructions"])

            # Address lookup tables
            address_table_lookups = TradeManager.parse_address_table_lookups(
                swap_instructions_data.get("addressLookupTableAddresses", [])
            )

            # Simulate to estimate compute units (only if Jupiter did NOT provide CU instructions)
            if not has_cu:
                blockhash_response = await wallet.client.get_latest_blockhash(commitment=Finalized)
                recent_blockhash = blockhash_response.value.blockhash

                msg = MessageV0.try_compile(
                    wallet.pubkey,
                    instructions,
                    [],
                    recent_blockhash,
                )
                transaction = VersionedTransaction.populate(msg, [Signature.default()])
                cu_units = (
                    await wallet.client.simulate_transaction(
                        transaction, sig_verify=False
                    )
                ).value.units_consumed or 1_400_000

                compute_budget_ix = set_compute_unit_limit(int(cu_units + 100_000))

                # Priority fee (helius)
                priority_fee_ix = None
                if provider == "helius":
                    priority_fee = swap_instructions_data.get("prioritizationFeeLamports", 0)
                    if priority_fee and priority_fee > 0:
                        priority_fee_ix = set_compute_unit_price(priority_fee)

                # Insert compute budget and priority fee at the start
                instructions_with_cu = [compute_budget_ix]
                if priority_fee_ix:
                    instructions_with_cu.append(priority_fee_ix)
                instructions_with_cu += instructions
            else:
                instructions_with_cu = instructions

            # Re-compile with compute budget and priority fee, now with address_table_lookups
            blockhash_response = await wallet.client.get_latest_blockhash(commitment=Finalized)
            recent_blockhash = blockhash_response.value.blockhash

            msg_final = MessageV0.try_compile(
                wallet.pubkey,
                instructions_with_cu,
                address_table_lookups,
                recent_blockhash,
            )

            # Sign and return the transaction
            if no_signer:
                signature = NullSigner(wallet.pubkey).sign_message(to_bytes_versioned(msg_final))
                signed_transaction = VersionedTransaction.populate(msg_final, [signature])
                return signed_transaction

            signature = wallet.sign_message(to_bytes_versioned(msg_final))
            signed_transaction = VersionedTransaction.populate(msg_final, [signature])
            return signed_transaction

        except Exception as e:
            logging.exception(f"Swap failed: {str(e)}")
            raise Exception(f"Swap failed: {str(e)}")